import logging

import numpy as np
import pytest
import xskillscore as xs

from climpred.exceptions import CoordinateError
from climpred.prediction import compute_hindcast


def test_same_inits_initializations(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that inits are identical at all leads for `same_inits` alignment."""
    with caplog.at_level(logging.INFO):
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="same_inits",
        )
        for i, record in enumerate(caplog.record_tuples):
            if i >= 2:
                print(record)
                assert "inits: 1954-01-01 00:00:00-2007-01-01 00:00:00" in record[2]


def test_same_inits_verification_dates(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that appropriate verifs are being used at each lead for `same_inits`
    alignment."""
    with caplog.at_level(logging.INFO):
        FIRST_INIT, LAST_INIT = 1954, 2007
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="same_inits",
        )
        nleads = hind_ds_initialized_1d_cftime["lead"].size
        for i, record in zip(
            np.arange(nleads + 2),
            caplog.record_tuples,
        ):
            if i >= 2:
                print(record)
                assert (
                    f"verifs: {FIRST_INIT+i}-01-01 00:00:00-{LAST_INIT+i}-01-01"
                    in record[2]
                )


@pytest.mark.parametrize("alignment", ["same_inits", "same_verifs"])
def test_disjoint_verif_time(small_initialized_da, small_verif_da, alignment):
    """Tests that alignment works with disjoint time in the verification
    data, i.e., non-continuous time sampling to verify against."""
    hind = small_initialized_da
    verif = small_verif_da.drop_sel(time=1992)
    actual = compute_hindcast(hind, verif, alignment=alignment, metric="mse")
    assert actual.notnull().all()
    # hindcast inits: [1990, 1991, 1992, 1993]
    # verif times: [1990, 1991, 1993, 1994]
    a = hind.sel(init=[1990, 1992, 1993]).rename({"init": "time"})
    b = verif.sel(time=[1991, 1993, 1994])
    a["time"] = b["time"]
    expected = xs.mse(a, b, "time")
    assert actual == expected


@pytest.mark.parametrize("alignment", ["same_inits", "same_verifs"])
def test_disjoint_inits(small_initialized_da, small_verif_da, alignment):
    """Tests that alignment works with disjoint inits in the verification
    data, i.e., non-continuous initializing to verify with."""
    hind = small_initialized_da.drop_sel(init=1991)
    verif = small_verif_da
    actual = compute_hindcast(hind, verif, alignment=alignment, metric="mse")
    assert actual.notnull().all()
    # hindcast inits: [1990, 1992, 1993]
    # verif times: [1990, 1991, 1992, 1993, 1994]
    a = hind.rename({"init": "time"})
    b = verif.sel(time=[1991, 1993, 1994])
    a["time"] = b["time"]
    expected = xs.mse(a, b, "time")
    assert actual == expected


def test_same_verifs_verification_dates(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that verifs are identical at all leads for `same_verifs` alignment."""
    with caplog.at_level(logging.INFO):
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="same_verifs",
        )
        for i, record in enumerate(caplog.record_tuples):
            if i >= 2:
                print(record)
                assert "verifs: 1964-01-01 00:00:00-2017-01-01 00:00:00" in record[2]


def test_same_verifs_initializations(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that appropriate verifs are being used at each lead for `same_inits`
    alignment."""
    with caplog.at_level(logging.INFO):
        FIRST_INIT, LAST_INIT = 1964, 2017
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="same_verifs",
        )
        nleads = hind_ds_initialized_1d_cftime["lead"].size
        for i, record in zip(
            np.arange(nleads + 2),
            caplog.record_tuples,
        ):
            if i >= 2:
                print(record)
                assert (
                    f"inits: {FIRST_INIT-i}-01-01 00:00:00-{LAST_INIT-i}-01-01 00:00:00"
                    in record[2]
                )


def test_same_verifs_raises_error_when_not_possible(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime
):
    """Tests that appropriate error is raised when a common set of verification dates
    cannot be found with the supplied initializations."""
    hind = hind_ds_initialized_1d_cftime.isel(lead=slice(0, 3), init=[1, 3, 5, 7, 9])
    with pytest.raises(CoordinateError):
        compute_hindcast(hind, reconstruction_ds_1d_cftime, alignment="same_verifs")


def test_maximize_alignment_inits(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that appropriate inits are selected for `maximize` alignment."""
    with caplog.at_level(logging.INFO):
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="maximize",
        )
        # Add dummy values for the first two lines since they are just metadata.
        for i, record in zip(
            np.concatenate(([0, 0], hind_ds_initialized_1d_cftime.lead.values)),
            caplog.record_tuples,
        ):
            if i >= 1:
                print(record)
                assert (
                    f"inits: 1954-01-01 00:00:00-{2016-i}-01-01 00:00:00" in record[2]
                )


def test_maximize_alignment_verifs(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that appropriate verifs are selected for `maximize` alignment."""
    with caplog.at_level(logging.INFO):
        compute_hindcast(
            hind_ds_initialized_1d_cftime,
            reconstruction_ds_1d_cftime,
            alignment="maximize",
        )
        # Add dummy values for the first two lines since they are just metadata.
        for i, record in zip(
            np.concatenate(([0, 0], hind_ds_initialized_1d_cftime.lead.values)),
            caplog.record_tuples,
        ):
            if i >= 1:
                print(record)
                assert (
                    f"verifs: {1955+i}-01-01 00:00:00-2017-01-01 00:00:00" in record[2]
                )



import numpy as np
import pytest
import xarray as xr
import xskillscore as xs
from xarray.testing import assert_allclose

from climpred.bootstrap import bootstrap_perfect_model
from climpred.comparisons import PM_COMPARISONS
from climpred.metrics import __ALL_METRICS__ as all_metrics, Metric, __pearson_r
from climpred.prediction import compute_hindcast, compute_perfect_model


def my_mse_function(forecast, verif, dim=None, **metric_kwargs):
    # function
    return ((forecast - verif) ** 2).mean(dim)


my_mse = Metric(
    name="mse",
    function=my_mse_function,
    positive=True,
    probabilistic=False,
    unit_power=2,
    long_name="MSE",
    aliases=["mSe", "<<<SE"],
)

ITERATIONS = 2


@pytest.mark.parametrize("comparison", PM_COMPARISONS)
def test_custom_metric_passed_to_compute(
    PM_da_initialized_1d, PM_da_control_1d, comparison
):
    """Test custom metric in compute_perfect_model."""
    actual = compute_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        comparison=comparison,
        metric=my_mse,
        dim="init",
    )

    expected = compute_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        comparison=comparison,
        metric="mse",
        dim="init",
    )

    assert_allclose(actual, expected)


@pytest.mark.slow
def test_custom_metric_passed_to_bootstrap_compute(
    PM_da_initialized_1d, PM_da_control_1d
):
    """Test custom metric in bootstrap_perfect_model."""
    comparison = "e2c"
    dim = "init"
    np.random.seed(42)
    actual = bootstrap_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        comparison=comparison,
        metric=my_mse,
        iterations=ITERATIONS,
        dim=dim,
    )

    expected = bootstrap_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        comparison=comparison,
        metric="mse",
        iterations=ITERATIONS,
        dim=dim,
    )

    assert_allclose(actual, expected, rtol=0.1, atol=1)


@pytest.mark.parametrize("metric", ("rmse", "mse"))
def test_pm_metric_skipna(PM_da_initialized_3d, PM_da_control_3d, metric):
    """Test skipna in compute_perfect_model."""
    PM_da_initialized_3d = PM_da_initialized_3d.copy()
    # manipulating data
    PM_da_initialized_3d.values[1:3, 1:4, 1:4, 4:6, 4:6] = np.nan

    base = compute_perfect_model(
        PM_da_initialized_3d,
        PM_da_control_3d,
        metric=metric,
        skipna=False,
        dim="init",
        comparison="m2e",
    ).mean("member")
    skipping = compute_perfect_model(
        PM_da_initialized_3d,
        PM_da_control_3d,
        metric=metric,
        skipna=True,
        dim="init",
        comparison="m2e",
    ).mean("member")
    assert ((base - skipping) != 0.0).any()
    assert base.isel(lead=2, x=5, y=5).isnull()
    assert not skipping.isel(lead=2, x=5, y=5).isnull()


@pytest.mark.skip(reason="comparisons dont work here")
@pytest.mark.parametrize("metric", ("rmse", "mse"))
@pytest.mark.parametrize("comparison", ["m2e", "m2m"])
def test_pm_metric_weights_m2x(
    PM_da_initialized_3d, PM_da_control_3d, comparison, metric
):
    """Test init weights in compute_perfect_model."""
    # distribute weights on initializations
    dim = "init"
    base = compute_perfect_model(
        PM_da_initialized_3d,
        PM_da_control_3d,
        dim=dim,
        metric=metric,
        comparison=comparison,
    )
    weights = xr.DataArray(np.arange(1, 1 + PM_da_initialized_3d[dim].size), dims=dim)
    weights = xr.DataArray(
        np.arange(
            1,
            1 + PM_da_initialized_3d[dim].size * PM_da_initialized_3d["member"].size,
        ),
        dims="init",
    )

    weighted = compute_perfect_model(
        PM_da_initialized_3d,
        PM_da_control_3d,
        dim=dim,
        comparison=comparison,
        metric=metric,
        weights=weights,
    )
    print((base / weighted).mean(["x", "y"]))
    # test for difference
    assert (xs.smape(base, weighted, ["x", "y"]) > 0.01).any()


@pytest.mark.parametrize("metric", ("rmse", "mse"))
def test_hindcast_metric_skipna(hind_da_initialized_3d, reconstruction_da_3d, metric):
    """Test skipna argument in hindcast_metric."""
    # manipulating data with nans
    hind_da_initialized_3d[0, 2, 0, 2] = np.nan
    base = compute_hindcast(
        hind_da_initialized_3d,
        reconstruction_da_3d,
        metric=metric,
        skipna=False,
        dim="init",
        alignment="same_inits",
    )
    skipping = compute_hindcast(
        hind_da_initialized_3d,
        reconstruction_da_3d,
        metric=metric,
        skipna=True,
        dim="init",
        alignment="same_inits",
    )
    div = base / skipping
    assert (div != 1).any()


@pytest.mark.skip(reason="comparisons dont work here")
@pytest.mark.parametrize("metric", ("rmse", "mse"))
@pytest.mark.parametrize("comparison", ["e2o", "m2o"])
def test_hindcast_metric_weights_x2r(
    hind_da_initialized_3d, reconstruction_da_3d, comparison, metric
):
    """Test init weights in compute_hindcast."""
    dim = "init"
    base = compute_hindcast(
        hind_da_initialized_3d,
        reconstruction_da_3d,
        dim=dim,
        metric=metric,
        comparison=comparison,
    )
    weights = xr.DataArray(np.arange(1, 1 + hind_da_initialized_3d[dim].size), dims=dim)
    weights = xr.DataArray(
        np.arange(
            1,
            1
            + hind_da_initialized_3d[dim].size * hind_da_initialized_3d["member"].size,
        ),
        dims="init",
    )

    weighted = compute_hindcast(
        hind_da_initialized_3d,
        reconstruction_da_3d,
        dim=dim,
        comparison=comparison,
        metric=metric,
        weights=weights,
    )
    print((base / weighted).mean(["nlon", "nlat"]))
    # test for difference
    assert (xs.smape(base, weighted, ["nlat", "nlon"]) > 0.01).any()


def test_Metric_display():
    summary = __pearson_r.__repr__()
    assert "Kind: deterministic" in summary.split("\n")[4]


def test_no_repeating_metric_aliases():
    """Tests that there are no repeating aliases for metrics, which would overwrite
    the earlier defined metric."""
    METRICS = []
    for m in all_metrics:
        if m.aliases is not None:
            for a in m.aliases:
                METRICS.append(a)
    duplicates = set([x for x in METRICS if METRICS.count(x) > 1])
    print(f"Duplicate metrics: {duplicates}")
    assert len(duplicates) == 0


def test_contingency(hindcast_hist_obs_1d):
    """Test contingency table perfect results."""
    hindcast = hindcast_hist_obs_1d
    hindcast = hindcast.map(xr.ones_like)
    category_edges = np.array([-0.5, 0.0, 0.5, 0.9, 1.1])
    metric_kwargs = {
        "forecast_category_edges": category_edges,
        "observation_category_edges": category_edges,
        "score": "table",
    }
    skill = hindcast.verify(
        metric="contingency",
        comparison="m2o",
        dim=["member", "init"],
        alignment="same_verifs",
        **metric_kwargs,
    ).SST

    assert skill.isel(observations_category=-1, forecasts_category=-1).notnull().all()
    assert (
        skill.isel(
            observations_category=slice(None, -1), forecasts_category=slice(None, -1)
        )
        == 0.0
    ).all()



from os.path import exists

from setuptools import find_packages, setup

if exists("README.rst"):
    with open("README.rst") as f:
        long_description = f.read()
else:
    long_description = ""

with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")

test_requirements = ["pytest-cov"]
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]

setup(
    maintainer="Riley X. Brady and Aaron Spring",
    maintainer_email="riley.brady@colorado.edu",
    description="Verification of weather and climate forecasts." + " prediction.",
    install_requires=install_requires,
    python_requires=">=3.7",
    license="MIT",
    long_description=long_description,
    classifiers=CLASSIFIERS,
    name="climpred",
    packages=find_packages(),
    test_suite="climpred/tests",
    tests_require=test_requirements,
    url="https://github.com/pangeo-data/climpred",
    use_scm_version={"version_scheme": "post-release", "local_scheme": "dirty-tag"},
    setup_requires=[
        "setuptools_scm",
        "setuptools>=30.3.0",
        "setuptools_scm_git_archive",
    ],
    zip_safe=False,
)



import logging
import warnings

import numpy as np
import pandas as pd
import xarray as xr
from bias_correction import XBiasCorrection
from xclim import sdba
from xskillscore.core.utils import suppress_warnings

from .constants import BIAS_CORRECTION_BIAS_CORRECTION_METHODS, GROUPBY_SEASONALITIES
from .metrics import Metric
from .options import OPTIONS
from .utils import (
    convert_cftime_to_datetime_coords,
    convert_time_index,
    get_lead_cftime_shift_args,
    shift_cftime_singular,
)


def sub(a, b):
    return a - b


def div(a, b):
    return a / b


def leave_one_out(bias, dim):
    """Leave-one-out creating a new dimension 'sample' and fill with np.NaN.

    See also: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.LeaveOneOut.html"""
    bias_nan = []
    for i in range(bias[dim].size):
        bias_nan.append(
            bias.drop_isel({dim: i}).reindex({dim: bias[dim]}).rename({dim: "sample"})
        )
    bias_nan = xr.concat(bias_nan, dim).assign_coords({dim: bias[dim]})
    return bias_nan


def leave_one_out_drop(bias, dim):
    """Leave-one-out creating a new dimension 'sample'.

    See also: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.LeaveOneOut.html."""
    bias_nan = []
    for i in range(bias[dim].size):
        bias_nan.append(bias.drop_isel({dim: i}).rename({dim: "sample"}).drop("sample"))
    bias_nan = xr.concat(bias_nan, dim).assign_coords(
        {dim: bias[dim], "sample": np.arange(bias[dim].size - 1)}
    )
    return bias_nan


def _mean_bias_removal_func(hind, bias, dim, how):
    """Quick removal of mean bias over all initializations without cross validation.

    Args:
        hind (xr.object): hindcast.
        bias (xr.object): bias.
        dim (str): Time dimension name in bias.

    Returns:
        xr.object: bias removed hind

    """
    how_operator = sub if how == "additive" else div
    seasonality = OPTIONS["seasonality"]

    with xr.set_options(keep_attrs=True):
        if seasonality == "weekofyear":
            # convert to datetime for weekofyear operations
            hind = convert_cftime_to_datetime_coords(hind, dim)
            bias = convert_cftime_to_datetime_coords(bias, dim)
        hind_groupby = f"{dim}.{seasonality}"

        bias_removed_hind = how_operator(
            hind.groupby(hind_groupby),
            bias.groupby(hind_groupby).mean(),
        )
    bias_removed_hind.attrs = hind.attrs
    # convert back to CFTimeIndex if needed
    if isinstance(bias_removed_hind[dim].to_index(), pd.DatetimeIndex):
        bias_removed_hind = convert_time_index(bias_removed_hind, dim, "hindcast")
    return bias_removed_hind


def _multiplicative_std_correction(hind, spread, dim, obs=None):
    """Quick removal of std bias over all initializations without cross validation.

    Args:
        hind (xr.object): hindcast.
        spread (xr.object): model spread.
        dim (str): Time dimension name in bias.
        obs (xr.object): observations

    Returns:
        xr.object: bias removed hind

    """
    seasonality = OPTIONS["seasonality"]
    if seasonality == "weekofyear":
        # convert to datetime for weekofyear operations
        hind = convert_cftime_to_datetime_coords(hind, "init")
        spread = convert_cftime_to_datetime_coords(spread, "init")
        obs = convert_cftime_to_datetime_coords(obs, "time")

    init_groupby = f"init.{seasonality}"
    obs_groupby = f"time.{seasonality}"

    with xr.set_options(keep_attrs=True):
        model_mean_spread = spread.groupby(init_groupby).mean()
        model_member_mean = hind.mean("member").groupby(init_groupby).mean()
        # assume that no trend here
        obs_spread = obs.groupby(obs_groupby).std()

        # z distr
        init_z = (hind.groupby(init_groupby) - model_member_mean).groupby(
            init_groupby
        ) / model_mean_spread

        # scale with obs_spread and model mean
        init_std_corrected = (init_z.groupby(init_groupby) * obs_spread).groupby(
            init_groupby
        ) + model_member_mean

    init_std_corrected.attrs = hind.attrs
    # convert back to CFTimeIndex if needed
    if isinstance(init_std_corrected.init.to_index(), pd.DatetimeIndex):
        init_std_corrected = convert_time_index(init_std_corrected, "init", "hindcast")
    return init_std_corrected


def _std_multiplicative_bias_removal_func_cv(hind, spread, dim, obs, cv="LOO"):
    """Remove std bias from all but the given initialization (cross-validation).

    .. note::
        This method follows Jolliffe 2011. For a given initialization, bias is computed
        over all other initializations, excluding the one in question. This calculated
        bias is removed from the given initialization, and then the process proceeds to
        the following one.

    Args:
        hind (xr.object): hindcast.
        bias (xr.object): bias.
        dim (str): Time dimension name in bias.
        how (str): additive or multiplicative bias.

    Returns:
        xr.object: bias removed hind

    Reference:
        * Jolliffe, Ian T., and David B. Stephenson. Forecast Verification: A
          Practitioner’s Guide in Atmospheric Science. Chichester, UK: John Wiley &
          Sons, Ltd, 2011. https://doi.org/10.1002/9781119960003., Chapter: 5.3.1, p.80
    """
    seasonality = OPTIONS["seasonality"]
    if seasonality == "weekofyear":
        # convert to datetime for weekofyear operations
        hind = convert_cftime_to_datetime_coords(hind, "init")
        spread = convert_cftime_to_datetime_coords(spread, "init")
        obs = convert_cftime_to_datetime_coords(obs, "time")

    bias_removed_hind = []
    for init in hind.init.data:
        hind_drop_init = hind.drop_sel(init=init).init

        with xr.set_options(keep_attrs=True):
            init_groupby = f"init.{seasonality}"
            time_groupby = f"time.{seasonality}"

            model_mean_spread = (
                spread.sel(init=hind_drop_init).groupby(init_groupby).mean()
            )
            model_member_mean = (
                hind.drop_sel(init=init).mean("member").groupby(init_groupby).mean()
            )
            # assume that no trend here
            obs_spread = obs.groupby(time_groupby).std()

            # z distr
            init_z = (
                hind.sel(init=[init]).groupby(init_groupby) - model_member_mean
            ).groupby(init_groupby) / model_mean_spread
            # scale with obs_spread and model mean
            init_std_corrected = (init_z.groupby(init_groupby) * obs_spread).groupby(
                init_groupby
            ) + model_member_mean

        bias_removed_hind.append(init_std_corrected)

    init_std_corrected = xr.concat(bias_removed_hind, "init")

    init_std_corrected.attrs = hind.attrs
    # convert back to CFTimeIndex if needed
    if isinstance(init_std_corrected.init.to_index(), pd.DatetimeIndex):
        init_std_corrected = convert_time_index(init_std_corrected, "init", "hindcast")
    return init_std_corrected


def _mean_bias_removal_func_cv(hind, bias, dim, how, cv="LOO"):
    """Remove mean bias from all but the given initialization (cross-validation).

    .. note::
        This method follows Jolliffe 2011. For a given initialization, bias is computed
        over all other initializations, excluding the one in question. This calculated
        bias is removed from the given initialization, and then the process proceeds to
        the following one.

    Args:
        hind (xr.object): hindcast.
        bias (xr.object): bias.
        dim (str): Time dimension name in bias.
        how (str): additive or multiplicative bias.

    Returns:
        xr.object: bias removed hind

    Reference:
        * Jolliffe, Ian T., and David B. Stephenson. Forecast Verification: A
          Practitioner’s Guide in Atmospheric Science. Chichester, UK: John Wiley &
          Sons, Ltd, 2011. https://doi.org/10.1002/9781119960003., Chapter: 5.3.1, p.80
    """
    how_operator = sub if how == "additive" else div
    seasonality = OPTIONS["seasonality"]
    bias = bias.rename({dim: "init"})
    bias_removed_hind = []
    logging.info(f"mean {how} bias removal with seasonality {seasonality}:")
    if seasonality == "weekofyear":
        # convert to datetime for weekofyear operations
        hind = convert_cftime_to_datetime_coords(hind, "init")
        bias = convert_cftime_to_datetime_coords(bias, "init")

    if cv == "LOO":
        for init in hind.init.data:
            hind_drop_init = hind.drop_sel(init=init).init
            with xr.set_options(keep_attrs=True):
                init_groupby = f"init.{seasonality}"
                init_bias_removed = how_operator(
                    hind.sel(init=[init]).groupby(init_groupby),
                    bias.sel(
                        init=hind_drop_init.to_index().intersection(
                            bias.init.to_index()
                        )
                    )
                    .groupby(init_groupby)
                    .mean(),
                )
                if seasonality in init_bias_removed.coords:
                    del init_bias_removed.coords[seasonality]
            bias_removed_hind.append(init_bias_removed)
        bias_removed_hind = xr.concat(bias_removed_hind, "init")
    else:
        raise NotImplementedError(f'try cv="LOO", found {cv}')
    bias_removed_hind.attrs = hind.attrs
    # convert back to CFTimeIndex if needed
    if isinstance(bias_removed_hind.init.to_index(), pd.DatetimeIndex):
        bias_removed_hind = convert_time_index(bias_removed_hind, "init", "hindcast")
    return bias_removed_hind


def gaussian_bias_removal(
    hindcast,
    alignment,
    cv=False,
    how="additive_mean",
    train_test_split="fair",
    train_time=None,
    train_init=None,
    **metric_kwargs,
):
    """Calc bias based on OPTIONS['seasonality'] and remove bias from py:class:`~climpred.classes.HindcastEnsemble`.

    Args:
        hindcast (HindcastEnsemble): hindcast.
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be
            based on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior
            to computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        how (str): what kind of bias removal to perform. Select
            from ['additive_mean', 'multiplicative_mean','multiplicative_std']. Defaults to 'additive_mean'.
        cv (bool or str): Defaults to True.

            - True: Use cross validation in bias removal function.
                This excludes the given initialization from the bias calculation.
            - 'LOO': see True
            - False: include the given initialization in the calculation, which
                is much faster and but yields similar skill with a large N of
                initializations.

    Returns:
        HindcastEnsemble: bias removed hindcast.

    """
    if train_test_split == "fair":
        hindcast_train = hindcast.copy()
        hindcast_test = hindcast.copy()
        if alignment in ["same_inits", "maximize"]:
            hindcast_train._datasets["initialized"] = hindcast.get_initialized().sel(
                init=train_init
            )  # for bias
            hindcast_test._datasets[
                "initialized"
            ] = hindcast.get_initialized().drop_sel(
                init=train_init
            )  # to reduce bias
        if alignment in ["same_verif"]:
            train_time = hindcast.coords["time"].sel(time=train_time).to_index()
            # add inits before lead.max()
            n, freq = get_lead_cftime_shift_args(
                hindcast.coords["lead"].attrs["units"], hindcast.coords["lead"].max()
            )
            train_time_init = train_time.union(train_time.shift(-n, freq)).intersection(
                hindcast.coords["init"].to_index()
            )
            hindcast_train._datasets["initialized"] = hindcast.get_initialized().sel(
                init=train_time_init
            )
            hindcast_test._datasets[
                "initialized"
            ] = hindcast.get_initialized().drop_sel(init=train_time_init)
    else:
        assert train_test_split in ["unfair", "unfair-cv"]
        hindcast_train = hindcast
        hindcast_test = hindcast

    if "mean" in how:
        # calculate bias lead-time dependent
        bias = hindcast_train.verify(
            metric="unconditional_bias" if how == "additive_mean" else "mul_bias",
            comparison="e2o",
            dim=[],  # not used by bias func, therefore best to add [] here
            alignment=alignment,
            **metric_kwargs,
        )

    if how == "multiplicative_std":
        bias = hindcast_train.verify(
            metric="spread",
            comparison="m2o",
            dim="member",
            alignment=alignment,
        )

    # how to remove bias
    if "mean" in how:
        if cv in [False, None]:
            bias_removal_func = _mean_bias_removal_func
            bias_removal_func_kwargs = dict(how=how.split("_")[0])
        else:
            bias_removal_func = _mean_bias_removal_func_cv
            bias_removal_func_kwargs = dict(how=how.split("_")[0], cv=cv)

    elif how == "multiplicative_std":
        if cv in [False, None]:
            bias_removal_func = _multiplicative_std_correction
            bias_removal_func_kwargs = dict(
                obs=hindcast.get_observations(),
            )
        else:
            bias_removal_func = _std_multiplicative_bias_removal_func_cv
            bias_removal_func_kwargs = dict(obs=hindcast.get_observations(), cv=cv)

    bias_removed_hind = bias_removal_func(
        hindcast_test.get_initialized(), bias, "init", **bias_removal_func_kwargs
    )
    bias_removed_hind = bias_removed_hind.squeeze(drop=True)

    # remove groupby label from coords
    for c in GROUPBY_SEASONALITIES + ["skill"]:
        if c in bias_removed_hind.coords and c not in bias_removed_hind.dims:
            del bias_removed_hind.coords[c]

    # replace raw with bias reducted initialized dataset
    hindcast_bias_removed = hindcast.copy()
    hindcast_bias_removed._datasets["initialized"] = bias_removed_hind
    return hindcast_bias_removed


def bias_correction(
    hindcast,
    alignment,
    cv=False,
    how="normal_mapping",
    train_test_split="fair",
    train_time=None,
    train_init=None,
    **metric_kwargs,
):
    """Calc bias based on OPTIONS['seasonality'] and remove bias from py:class:`~climpred.classes.HindcastEnsemble`.

    Args:
        hindcast (HindcastEnsemble): hindcast.
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be
            based on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior
            to computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        how (str): what kind of bias removal to perform. Select
            from ['additive_mean', 'multiplicative_mean','multiplicative_std']. Defaults to 'additive_mean'.
        cv (bool): Use cross validation in bias removal function. This
            excludes the given initialization from the bias calculation. With False,
            include the given initialization in the calculation, which is much faster
            but yields similar skill with a large N of initializations.
            Defaults to True.

    Returns:
        HindcastEnsemble: bias removed hindcast.

    """

    def bc_func(
        forecast,
        observations,
        dim=None,
        method=how,
        cv=False,
        **metric_kwargs,
    ):
        """Wrapping https://github.com/pankajkarman/bias_correction/blob/master/bias_correction.py.

        Functions to perform bias correction of datasets to remove biases across datasets. Implemented methods include:
        - quantile mapping: https://rmets.onlinelibrary.wiley.com/doi/pdf/10.1002/joc.2168)
        - modified quantile mapping: https://www.sciencedirect.com/science/article/abs/pii/S0034425716302000?via%3Dihub
        - scaled distribution mapping (Gamma and Normal Corrections): https://www.hydrol-earth-syst-sci.net/21/2649/2017/
        """
        corrected = []
        seasonality = OPTIONS["seasonality"]
        dim = "time"
        if seasonality == "weekofyear":
            forecast = convert_cftime_to_datetime_coords(forecast, dim)
            observations = convert_cftime_to_datetime_coords(observations, dim)

        if train_test_split in ["fair"]:
            if alignment in ["same_inits", "maximize"]:
                train_dim = train_init.rename({"init": "time"})
                # shift init to time
                n, freq = get_lead_cftime_shift_args(
                    forecast.lead.attrs["units"], forecast.lead
                )
                train_dim = shift_cftime_singular(train_dim[dim], n, freq)
                data_to_be_corrected = forecast.drop_sel({dim: train_dim})
            elif alignment in ["same_verif"]:
                train_dim = train_time
                intersection = (
                    train_dim[dim].to_index().intersection(forecast[dim].to_index())
                )
                data_to_be_corrected = forecast.drop_sel({dim: intersection})

            intersection = (
                train_dim[dim].to_index().intersection(forecast[dim].to_index())
            )
            forecast = forecast.sel({dim: intersection})
            reference = observations.sel({dim: intersection})
        else:
            model = forecast
            data_to_be_corrected = forecast
            reference = observations

        data_to_be_corrected_ori = data_to_be_corrected.copy()

        for label, group in forecast.groupby(f"{dim}.{seasonality}"):
            reference = observations.sel({dim: group[dim]})
            model = forecast.sel({dim: group[dim]})
            if train_test_split in ["unfair", "unfair-cv"]:
                # take all
                data_to_be_corrected = forecast.sel({dim: group[dim]})
            else:
                group_dim_data_to_be_corrected = (
                    getattr(data_to_be_corrected_ori[dim].dt, seasonality) == label
                )
                data_to_be_corrected = data_to_be_corrected_ori.sel(
                    {dim: group_dim_data_to_be_corrected}
                )

            if cv == "LOO" and train_test_split == "unfair-cv":
                reference = leave_one_out(reference, dim)
                model = leave_one_out(model, dim)
                data_to_be_corrected = leave_one_out(data_to_be_corrected, dim)

            dim2 = "time_member"
            if "member" in model.dims:
                reference = reference.broadcast_like(model)
                data_to_be_corrected = data_to_be_corrected.broadcast_like(model)
                model = model.stack({dim2: ["time", "member"]})
                reference = reference.stack({dim2: ["time", "member"]})
                data_to_be_corrected = data_to_be_corrected.stack(
                    {dim2: ["time", "member"]}
                )
            dim_used = dim2 if "member" in forecast.dims else dim

            # using bias-correction: https://github.com/pankajkarman/bias_correction/blob/master/bias_correction.py
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                bc = XBiasCorrection(
                    reference,
                    model,
                    data_to_be_corrected,
                    dim=dim_used,
                )
                c = bc.correct(method=method, join="outer", **metric_kwargs)
            if dim2 in c.dims:
                c = c.unstack(dim2)
            if cv and dim in c.dims and "sample" in c.dims:
                c = c.mean(dim)
                c = c.rename({"sample": dim})
            # select only where data_to_be_corrected was input
            if dim2 in data_to_be_corrected.dims:
                data_to_be_corrected = data_to_be_corrected.unstack(dim2)
            c = c.sel({dim: data_to_be_corrected[dim]})
            corrected.append(c)
        corrected = xr.concat(corrected, dim).sortby(dim)
        # convert back to CFTimeIndex if needed
        if isinstance(corrected[dim].to_index(), pd.DatetimeIndex):
            corrected = convert_time_index(corrected, dim, "hindcast")
        return corrected

    bc = Metric(
        "bias_correction", bc_func, positive=False, probabilistic=False, unit_power=1
    )

    # calculate bias lead-time dependent
    bias_removed_hind = hindcast.verify(
        metric=bc,
        comparison="m2o" if "member" in hindcast.dims else "e2o",
        dim=[],  # set internally inside bc
        alignment=alignment,
        cv=cv,
        **metric_kwargs,
    ).squeeze(drop=True)

    # remove groupby label from coords
    for c in GROUPBY_SEASONALITIES + ["skill"]:
        if c in bias_removed_hind.coords and c not in bias_removed_hind.dims:
            del bias_removed_hind.coords[c]

    # keep attrs
    bias_removed_hind.attrs = hindcast.get_initialized().attrs
    for v in bias_removed_hind.data_vars:
        bias_removed_hind[v].attrs = hindcast.get_initialized()[v].attrs

    # replace raw with bias reducted initialized dataset
    hindcast_bias_removed = hindcast.copy()
    hindcast_bias_removed._datasets["initialized"] = bias_removed_hind

    return hindcast_bias_removed


def xclim_sdba(
    hindcast,
    alignment,
    cv=False,
    how="DetrendedQuantileMapping",
    train_test_split="fair",
    train_time=None,
    train_init=None,
    **metric_kwargs,
):
    """Calc bias based on grouper to be passed as metric_kwargs and remove bias from py:class:`~climpred.classes.HindcastEnsemble`.

    See climpred.constants.XCLIM_BIAS_CORRECTION_METHODS for implemented methods for ``how``.

    Args:
        hindcast (HindcastEnsemble): hindcast.
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be
            based on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior
            to computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        how (str): not used
        cv (bool): Use cross validation in removal function. This
            excludes the given initialization from the bias calculation. With False,
            include the given initialization in the calculation, which is much faster
            but yields similar skill with a large N of initializations.
            Defaults to True.

    Returns:
        HindcastEnsemble: bias removed hindcast.

    """

    def bc_func(
        forecast,
        observations,
        dim=None,
        method=how,
        cv=False,
        **metric_kwargs,
    ):
        """Wrapping https://github.com/Ouranosinc/xclim/blob/master/xclim/sdba/adjustment.py.

        Functions to perform bias correction of datasets to remove biases across datasets. See climpred.constants.XCLIM_BIAS_CORRECTION_METHODS for implemented methods.
        """
        seasonality = OPTIONS["seasonality"]
        dim = "time"
        if seasonality == "weekofyear":
            forecast = convert_cftime_to_datetime_coords(forecast, dim)
            observations = convert_cftime_to_datetime_coords(observations, dim)

        if train_test_split in ["fair"]:
            if alignment in ["same_inits", "maximize"]:
                train_dim = train_init.rename({"init": "time"})
                # shift init to time
                n, freq = get_lead_cftime_shift_args(
                    forecast.lead.attrs["units"], forecast.lead
                )
                train_dim = shift_cftime_singular(train_dim[dim], n, freq)
                data_to_be_corrected = forecast.drop_sel({dim: train_dim})
            elif alignment in ["same_verif"]:
                train_dim = train_time
                intersection = (
                    train_dim[dim].to_index().intersection(forecast[dim].to_index())
                )
                data_to_be_corrected = forecast.drop_sel({dim: intersection})

            intersection = (
                train_dim[dim].to_index().intersection(forecast[dim].to_index())
            )
            forecast = forecast.sel({dim: intersection})
            model = forecast
            reference = observations.sel({dim: intersection})
        else:
            model = forecast
            data_to_be_corrected = forecast
            reference = observations

        if train_test_split in ["unfair", "unfair-cv"]:
            # take all
            data_to_be_corrected = forecast

        if cv == "LOO" and train_test_split == "unfair-cv":
            reference = leave_one_out(reference, dim)
            model = leave_one_out(model, dim)
            data_to_be_corrected = leave_one_out(data_to_be_corrected, dim)

        if "group" not in metric_kwargs:
            metric_kwargs["group"] = dim + "." + OPTIONS["seasonality"]
        elif metric_kwargs["group"] is None:
            metric_kwargs["group"] = dim + "." + OPTIONS["seasonality"]

        if "init" in metric_kwargs["group"]:
            metric_kwargs["group"] = metric_kwargs["group"].replace("init", "time")
        if "member" in model.dims:
            metric_kwargs["add_dims"] = ["member"]
            if "member" not in reference.dims:
                reference = reference.expand_dims(member=[model.member[0]])

        adjust_kwargs = {}
        for k in ["interp", "extrapolation", "detrend"]:
            if k in metric_kwargs:
                adjust_kwargs[k] = metric_kwargs.pop(k)

        def adjustment(reference, model, data_to_be_corrected):
            dqm = getattr(sdba.adjustment, method).train(
                reference, model, **metric_kwargs
            )
            return dqm.adjust(data_to_be_corrected, **adjust_kwargs)

        del model.coords["lead"]

        c = xr.Dataset()
        for v in model.data_vars:
            c[v] = adjustment(reference[v], model[v], data_to_be_corrected[v])

        if cv and dim in c.dims and "sample" in c.dims:
            c = c.mean(dim)
            c = c.rename({"sample": dim})
        # select only where data_to_be_corrected was input
        corrected = c.sel({dim: data_to_be_corrected[dim]})
        # convert back to CFTimeIndex if needed
        if isinstance(corrected[dim].to_index(), pd.DatetimeIndex):
            corrected = convert_time_index(corrected, dim, "hindcast")
        return corrected

    bc = Metric(
        "bias_correction", bc_func, positive=False, probabilistic=False, unit_power=1
    )

    # calculate bias lead-time dependent
    bias_removed_hind = hindcast.verify(
        metric=bc,
        comparison="m2o" if "member" in hindcast.dims else "e2o",
        dim=[],  # set internally inside bc
        alignment=alignment,
        cv=cv,
        **metric_kwargs,
    ).squeeze(drop=True)

    # remove groupby label from coords
    for c in GROUPBY_SEASONALITIES + ["skill"]:
        if c in bias_removed_hind.coords and c not in bias_removed_hind.dims:
            del bias_removed_hind.coords[c]

    # keep attrs
    bias_removed_hind.attrs = hindcast.get_initialized().attrs
    for v in bias_removed_hind.data_vars:
        bias_removed_hind[v].attrs = hindcast.get_initialized()[v].attrs

    # replace raw with bias reducted initialized dataset
    hindcast_bias_removed = hindcast.copy()
    hindcast_bias_removed._datasets["initialized"] = bias_removed_hind

    return hindcast_bias_removed



import numpy as np
import pytest
import xarray as xr
from xarray.testing import assert_equal

from climpred.comparisons import (
    __ALL_COMPARISONS__ as all_comparisons,
    PM_COMPARISONS,
    PROBABILISTIC_PM_COMPARISONS,
    Comparison,
    __e2c,
    __m2c,
    __m2e,
    __m2m,
)
from climpred.metrics import PM_METRICS, __mse as metric
from climpred.prediction import compute_perfect_model
from climpred.utils import get_comparison_class, get_metric_class


def test_e2c(PM_da_initialized_1d):
    """Test ensemble_mean-to-control (which can be any other one member) (e2c)
    comparison basic functionality.

    Clean comparison: Remove one control member from ensemble to use as reference.
    Take the remaining member mean as forecasts."""
    ds = PM_da_initialized_1d
    aforecast, areference = __e2c.function(ds, metric=metric)

    control_member = ds.member.values[0]
    reference = ds.sel(member=control_member, drop=True)
    # drop the member being reference
    ds = ds.drop_sel(member=control_member)
    forecast = ds.mean("member")

    eforecast, ereference = forecast, reference
    # very weak testing on shape
    assert eforecast.size == aforecast.size
    assert ereference.size == areference.size

    assert_equal(eforecast, aforecast)
    assert_equal(ereference, areference)


def test_m2c(PM_da_initialized_1d):
    """Test many-to-control (which can be any other one member) (m2c) comparison basic
    functionality.

    Clean comparison: Remove one control member from ensemble to use as reference.
    Take the remaining members as forecasts."""
    ds = PM_da_initialized_1d
    aforecast, areference = __m2c.function(ds, metric=metric)

    control_member = ds.member.values[0]
    reference = ds.sel(member=control_member, drop=True)
    # drop the member being reference
    ds_dropped = ds.drop_sel(member=control_member)
    forecast, reference = xr.broadcast(ds_dropped, reference)

    eforecast, ereference = forecast, reference
    # very weak testing on shape
    assert eforecast.size == aforecast.size
    assert ereference.size == areference.size

    assert_equal(eforecast, aforecast)
    assert_equal(ereference, areference)


def test_m2e(PM_da_initialized_1d):
    """Test many-to-ensemble-mean (m2e) comparison basic functionality.

    Clean comparison: Remove one member from ensemble to use as reference.
    Take the remaining members as forecasts."""
    ds = PM_da_initialized_1d
    aforecast, areference = __m2e.function(ds, metric=metric)

    reference_list = []
    forecast_list = []
    for m in ds.member.values:
        forecast = ds.drop_sel(member=m).mean("member")
        reference = ds.sel(member=m, drop=True)
        forecast, reference = xr.broadcast(forecast, reference)
        forecast_list.append(forecast)
        reference_list.append(reference)
    reference = xr.concat(reference_list, "member")
    forecast = xr.concat(forecast_list, "member")
    forecast["member"] = np.arange(forecast.member.size)
    reference["member"] = np.arange(reference.member.size)

    eforecast, ereference = forecast, reference
    # very weak testing on shape
    assert eforecast.size == aforecast.size
    assert ereference.size == areference.size

    assert_equal(eforecast, aforecast)
    assert_equal(ereference, areference)


def test_m2m(PM_da_initialized_1d):
    """Test many-to-many (m2m) comparison basic functionality.

    Clean comparison: Remove one member from ensemble to use as reference. Take the
    remaining members as forecasts."""
    ds = PM_da_initialized_1d
    aforecast, areference = __m2m.function(ds, metric=metric)

    reference_list = []
    forecast_list = []
    for m in ds.member.values:
        forecast = ds.drop_sel(member=m)
        forecast["member"] = np.arange(1, 1 + forecast.member.size)
        reference = ds.sel(member=m, drop=True)
        forecast, reference = xr.broadcast(forecast, reference)
        reference_list.append(reference)
        forecast_list.append(forecast)
    supervector_dim = "forecast_member"
    reference = xr.concat(reference_list, supervector_dim)
    forecast = xr.concat(forecast_list, supervector_dim)
    reference[supervector_dim] = np.arange(reference[supervector_dim].size)
    forecast[supervector_dim] = np.arange(forecast[supervector_dim].size)
    eforecast, ereference = forecast, reference
    # very weak testing here
    assert eforecast.size == aforecast.size
    assert ereference.size == areference.size


@pytest.mark.parametrize("metric", ["crps", "mse"])
@pytest.mark.parametrize("comparison", PM_COMPARISONS)
def test_all(PM_da_initialized_1d, comparison, metric):
    metric = get_metric_class(metric, PM_METRICS)
    ds = PM_da_initialized_1d
    comparison = get_comparison_class(comparison, PM_COMPARISONS)
    forecast, obs = comparison.function(ds, metric=metric)
    assert not forecast.isnull().any()
    assert not obs.isnull().any()
    if not metric.probabilistic:
        # same dimensions for deterministic metrics
        assert forecast.dims == obs.dims
    else:
        if comparison.name in PROBABILISTIC_PM_COMPARISONS:
            # same but member dim for probabilistic
            assert set(forecast.dims) - set(["member"]) == set(obs.dims)


def my_m2me_comparison(ds, metric=None):
    """Identical to m2e but median."""
    reference_list = []
    forecast_list = []
    supervector_dim = "member"
    for m in ds.member.values:
        forecast = ds.drop_sel(member=m).median("member")
        reference = ds.sel(member=m).squeeze()
        forecast_list.append(forecast)
        reference_list.append(reference)
    reference = xr.concat(reference_list, supervector_dim)
    forecast = xr.concat(forecast_list, supervector_dim)
    forecast[supervector_dim] = np.arange(forecast[supervector_dim].size)
    reference[supervector_dim] = np.arange(reference[supervector_dim].size)
    return forecast, reference


my_m2me_comparison = Comparison(
    name="m2me",
    function=my_m2me_comparison,
    probabilistic=False,
    hindcast=False,
)


@pytest.mark.parametrize("metric", ("rmse", "pearson_r"))
def test_new_comparison_passed_to_compute(
    PM_da_initialized_1d, PM_da_control_1d, metric
):
    actual = compute_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        comparison=my_m2me_comparison,
        metric=metric,
    )

    expected = compute_perfect_model(
        PM_da_initialized_1d, PM_da_control_1d, comparison="m2e", metric="mse"
    )

    assert (actual - expected).mean() != 0


def test_Comparison_display():
    summary = __m2m.__repr__()
    assert "Kind: deterministic and probabilistic" in summary.split("\n")[2]


def test_no_repeating_comparison_aliases():
    """Tests that there are no repeating aliases for comparison, which would overwrite
    the earlier defined comparison."""
    COMPARISONS = []
    for c in all_comparisons:
        if c.aliases is not None:
            for a in c.aliases:
                COMPARISONS.append(a)
    duplicates = set([x for x in COMPARISONS if COMPARISONS.count(x) > 1])
    print(f"Duplicate comparisons: {duplicates}")
    assert len(duplicates) == 0



# flake8: noqa
from pkg_resources import DistributionNotFound, get_distribution

from . import (
    bias_removal,
    bootstrap,
    comparisons,
    constants,
    exceptions,
    graphics,
    horizon,
    metrics,
    prediction,
    relative_entropy,
    smoothing,
    stats,
    testing,
    tutorial,
)
from .classes import HindcastEnsemble, PerfectModelEnsemble
from .options import set_options
from .preprocessing import mpi, shared
from .versioning.print_versions import show_versions

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass  # pragma: no cover



import pytest
import xarray as xr
from xarray.testing import assert_allclose

from climpred.bootstrap import bootstrap_hindcast
from climpred.comparisons import (
    PM_COMPARISONS,
    PROBABILISTIC_HINDCAST_COMPARISONS,
    PROBABILISTIC_PM_COMPARISONS,
)
from climpred.exceptions import DimensionError
from climpred.metrics import PM_METRICS
from climpred.utils import get_comparison_class, get_metric_class

xr.set_options(display_style="text")

ITERATIONS = 2

comparison_dim_PM = [
    ("m2m", "init"),
    ("m2m", "member"),
    ("m2m", ["init", "member"]),
    ("m2e", "init"),
    ("m2e", "member"),
    ("m2e", ["init", "member"]),
    ("m2c", "init"),
    ("m2c", "member"),
    ("m2c", ["init", "member"]),
    ("e2c", "init"),
]


@pytest.mark.parametrize("metric", ["crps", "mse"])
@pytest.mark.parametrize("comparison", PROBABILISTIC_PM_COMPARISONS)
def test_pm_comparison_stack_dims_when_deterministic(
    PM_da_initialized_1d, comparison, metric
):
    metric = get_metric_class(metric, PM_METRICS)
    comparison = get_comparison_class(comparison, PM_COMPARISONS)
    actual_f, actual_r = comparison.function(PM_da_initialized_1d, metric=metric)
    if not metric.probabilistic:
        assert "member" in actual_f.dims
        assert "member" in actual_r.dims
    else:
        assert "member" in actual_f.dims
        assert "member" not in actual_r.dims


# cannot work for e2c, m2e comparison because only 1:1 comparison
@pytest.mark.parametrize("comparison", PROBABILISTIC_PM_COMPARISONS)
def test_compute_perfect_model_dim_over_member(
    perfectModelEnsemble_initialized_control, comparison
):
    """Test deterministic metric calc skill over member dim."""
    actual = perfectModelEnsemble_initialized_control.verify(
        comparison=comparison,
        metric="rmse",
        dim="member",
    )["tos"]
    assert "init" in actual.dims
    assert not actual.isnull().any()
    # check that init is cftime object
    assert "cftime" in str(type(actual.init.values[0]))


# cannot work for e2o comparison because only 1:1 comparison
@pytest.mark.parametrize("comparison", PROBABILISTIC_HINDCAST_COMPARISONS)
def test_compute_hindcast_dim_over_member(hindcast_hist_obs_1d, comparison):
    """Test deterministic metric calc skill over member dim."""
    print(hindcast_hist_obs_1d.get_initialized().coords)
    actual = hindcast_hist_obs_1d.verify(
        comparison=comparison, metric="rmse", dim="member", alignment="same_verif"
    )["SST"]
    assert "init" in actual.dims
    # mean init because skill has still coords for init lead
    assert not actual.mean("init").isnull().any()


def test_compute_perfect_model_different_dims_quite_close(
    perfectModelEnsemble_initialized_control,
):
    """Tests nearly equal dim=['init','member'] and dim='member'."""
    stack_dims_true = perfectModelEnsemble_initialized_control.verify(
        comparison="m2c",
        metric="rmse",
        dim=["init", "member"],
    )["tos"]
    stack_dims_false = perfectModelEnsemble_initialized_control.verify(
        comparison="m2c",
        metric="rmse",
        dim="member",
    ).mean(["init"])["tos"]
    # no more than 10% difference
    assert_allclose(stack_dims_true, stack_dims_false, rtol=0.1, atol=0.03)


@pytest.mark.parametrize("metric", ["rmse", "pearson_r"])
@pytest.mark.parametrize(
    "comparison,dim",
    comparison_dim_PM,
)
def test_compute_pm_dims(
    perfectModelEnsemble_initialized_control, dim, comparison, metric
):
    """Test whether compute_pm calcs skill over many possible dims
    and comparisons and just reduces the result by dim."""
    print(dim)
    pm = perfectModelEnsemble_initialized_control
    actual = pm.verify(metric=metric, comparison=comparison, dim=dim)["tos"]
    if isinstance(dim, str):
        dim = [dim]
    # check whether only dim got reduced from coords
    if comparison == "e2c":  # dont expect member, remove manually
        assert set(pm.get_initialized().dims) - set(["member"]) - set(dim) == set(
            actual.dims
        ), print(pm.get_initialized().dims, "-", dim, "!=", actual.dims)
    else:
        assert set(pm.get_initialized().dims) - set(dim) == set(actual.dims), print(
            pm.get_initialized().dims, "-", dim, "!=", actual.dims
        )
    # check whether all nan
    if metric not in ["pearson_r"]:
        assert not actual.isnull().any()


@pytest.mark.parametrize(
    "metric,dim", [("rmse", "init"), ("rmse", "member"), ("crps", "member")]
)
def test_compute_hindcast_dims(hindcast_hist_obs_1d, dim, metric):
    """Test whether compute_hindcast calcs skill over all possible dims
    and comparisons and just reduces the result by dim."""
    actual = hindcast_hist_obs_1d.verify(
        metric=metric, dim=dim, comparison="m2o", alignment="same_verif"
    )["SST"]
    # check whether only dim got reduced from coords
    assert set(hindcast_hist_obs_1d.get_initialized().dims) - set(actual.dims) == set(
        [dim]
    )
    # check whether all nan
    if "init" in actual.dims:
        actual = actual.mean("init")
    assert not actual.isnull().any()


@pytest.mark.parametrize(
    "dim",
    ["init", "member", None, ["init", "member"], ["x", "y"], ["x", "y", "member"]],
)
def test_PM_multiple_dims(
    perfectModelEnsemble_initialized_control_3d_North_Atlantic, dim
):
    """Test that PerfectModelEnsemble accepts dims as subset from initialized dims."""
    pm = perfectModelEnsemble_initialized_control_3d_North_Atlantic
    assert pm.verify(metric="rmse", comparison="m2e", dim=dim).any()


def test_PM_multiple_dims_fail_if_not_in_initialized(
    perfectModelEnsemble_initialized_control_3d_North_Atlantic,
):
    """Test that PerfectModelEnsemble.verify() for multiple dims fails when not subset
    from initialized dims."""
    pm = perfectModelEnsemble_initialized_control_3d_North_Atlantic.isel(x=4, y=4)
    with pytest.raises(DimensionError) as excinfo:
        pm.verify(metric="rmse", comparison="m2e", dim=["init", "member", "x"])
    assert "is expected to be a subset of `initialized.dims`" in str(excinfo.value)


def test_PM_fails_probabilistic_member_not_in_dim(
    perfectModelEnsemble_initialized_control_3d_North_Atlantic,
):
    """Test that PerfectModelEnsemble.verify() raises ValueError for `member` not in
    dim if probabilistic metric."""
    pm = perfectModelEnsemble_initialized_control_3d_North_Atlantic
    with pytest.raises(ValueError) as excinfo:
        pm.verify(metric="crps", comparison="m2c", dim=["init"])
    assert (
        "requires to be computed over dimension `member`, which is not found in"
        in str(excinfo.value)
    )


@pytest.mark.parametrize("dim", [["member"], ["member", "init"]])
@pytest.mark.parametrize("metric", ["crpss", "crpss_es"])
def test_hindcast_crpss(hindcast_recon_1d_ym, metric, dim):
    """Test that CRPSS metrics reduce by dimension dim in HindcastEnsemble.verify()."""
    he = hindcast_recon_1d_ym.isel(lead=[0, 1])
    actual = he.verify(
        dim=dim, metric=metric, alignment="same_verif", comparison="m2o"
    ).squeeze()
    before_dims = he.get_initialized().dims
    debug_message = f"{before_dims} - {dim} != {actual.dims} but should be"
    for d in before_dims:
        if d not in dim:
            assert d in actual.dims, debug_message
        else:
            assert d not in actual.dims, debug_message


@pytest.mark.parametrize("comparison", ["m2m", "m2c"])
@pytest.mark.parametrize("dim", [["member"], ["member", "init"]])
@pytest.mark.parametrize("metric", ["crpss", "crpss_es"])
def test_pm_crpss(
    perfectModelEnsemble_initialized_control_1d_ym_cftime, metric, dim, comparison
):
    """Test that CRPSS metrics reduce by dimension dim in
    PerfectModelEnsemble.verify()."""
    pm = perfectModelEnsemble_initialized_control_1d_ym_cftime
    actual = pm.verify(dim=dim, metric=metric, comparison=comparison).squeeze()
    before_dims = pm.get_initialized().dims
    debug_message = f"{before_dims} - {dim} != {actual.dims} but should be"
    for d in before_dims:
        if d in dim:
            assert d not in actual.dims, debug_message
        else:
            assert d in actual.dims, debug_message


def test_pm_metric_weights(perfectModelEnsemble_initialized_control_3d_North_Atlantic):
    """Test PerfectModelEnsemble.verify() with weights yields different results."""
    pm = perfectModelEnsemble_initialized_control_3d_North_Atlantic
    skipna = True
    metric = "rmse"
    comparison = "m2e"
    dim = ["x", "y"]
    weights = pm.get_initialized()["lat"]
    s_no_weights = pm.verify(
        metric=metric, comparison=comparison, dim=dim, skipna=skipna
    )
    s_weights = pm.verify(
        metric=metric, comparison=comparison, dim=dim, skipna=skipna, weights=weights
    )
    # want to test for non equalness
    assert ((s_no_weights["tos"] - s_weights["tos"]) != 0).all()


def test_hindcast_metric_weights(hindcast_recon_3d):
    """Test HindcastEnsemble.verify() with weights yields different results."""
    he = hindcast_recon_3d
    skipna = True
    metric = "rmse"
    comparison = "e2o"
    dim = ["nlat", "nlon"]
    alignment = "same_verifs"
    weights = he.get_initialized()["TAREA"]
    s_no_weights = he.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        skipna=skipna,
        alignment=alignment,
    )
    s_weights = he.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        skipna=skipna,
        weights=weights,
        alignment=alignment,
    )
    # want to test for non equalness
    assert ((s_no_weights["SST"] - s_weights["SST"]) != 0).all()



import inspect

import xarray as xr

from .alignment import return_inits_and_verif_dates
from .checks import has_valid_lead_units, is_xarray
from .comparisons import (
    ALL_COMPARISONS,
    COMPARISON_ALIASES,
    HINDCAST_COMPARISONS,
    PM_COMPARISONS,
    __e2c,
)
from .constants import CLIMPRED_DIMS, M2M_MEMBER_DIM
from .metrics import (
    ALL_METRICS,
    DETERMINISTIC_HINDCAST_METRICS,
    METRIC_ALIASES,
    PM_METRICS,
    _rename_dim,
)
from .options import OPTIONS
from .utils import (
    assign_attrs,
    convert_time_index,
    get_comparison_class,
    get_lead_cftime_shift_args,
    get_metric_class,
    shift_cftime_index,
)


def persistence(verif, inits, verif_dates, lead):
    lforecast = verif.where(verif.time.isin(inits[lead]), drop=True)
    lverif = verif.sel(time=verif_dates[lead])
    return lforecast, lverif


def climatology(verif, inits, verif_dates, lead):
    seasonality_str = OPTIONS["seasonality"]
    if seasonality_str == "weekofyear":
        raise NotImplementedError
    climatology_day = verif.groupby(f"time.{seasonality_str}").mean()
    # enlarge times to get climatology_forecast times
    # this prevents errors if verification.time and hindcast.init are too much apart
    verif_hind_union = xr.DataArray(
        verif.time.to_index().union(inits[lead].time.to_index()), dims="time"
    )

    climatology_forecast = climatology_day.sel(
        {seasonality_str: getattr(verif_hind_union.time.dt, seasonality_str)},
        method="nearest",
    ).drop(seasonality_str)

    lforecast = climatology_forecast.where(
        climatology_forecast.time.isin(inits[lead]), drop=True
    )
    lverif = verif.sel(time=verif_dates[lead])
    return lforecast, lverif


def uninitialized(hist, verif, verif_dates, lead):
    """also called historical in some communities."""
    lforecast = hist.sel(time=verif_dates[lead])
    lverif = verif.sel(time=verif_dates[lead])
    return lforecast, lverif


# needed for PerfectModelEnsemble.verify(reference=...) and PredictionEnsemble.bootstrap
# TODO: should be refactored for better non-functional use within verify and bootstrap


def _adapt_member_for_reference_forecast(lforecast, lverif, metric, comparison, dim):
    """Maybe drop member from dim or add single-member dimension. Used in reference forecasts: climatology, uninitialized, persistence."""
    # persistence or climatology forecasts wont have member dimension, create if required
    # some metrics dont allow member dimension, remove and try mean
    # delete member from dim if needed
    if "member" in dim:
        if (
            "member" in lforecast.dims
            and "member" not in lverif.dims
            and not metric.requires_member_dim
        ):
            dim = dim.copy()
            dim.remove("member")
        elif "member" not in lforecast.dims and "member" not in lverif.dims:
            dim = dim.copy()
            dim.remove("member")
    # for probabilistic metrics requiring member dim, add single-member dimension
    if metric.requires_member_dim:
        if "member" not in lforecast.dims:
            lforecast = lforecast.expand_dims("member")  # add fake member dim
            if "member" not in dim:
                dim = dim.copy()
                dim.append("member")
        assert "member" in lforecast.dims and "member" not in lverif.dims
    # member not required by metric and not in dim but present in forecast
    if not metric.requires_member_dim:
        if "member" in lforecast.dims and "member" not in dim:
            lforecast = lforecast.mean("member")
    # multi member comparisons expect member dim
    if (
        comparison.name in ["m2o", "m2m", "m2c", "m2e"]
        and "member" not in lforecast.dims
        and metric.requires_member_dim
    ):
        lforecast = lforecast.expand_dims("member")  # add fake member dim
    return lforecast, dim


def compute_climatology(
    hind,
    verif=None,
    metric="pearson_r",
    comparison="m2e",
    alignment="same_inits",
    add_attrs=True,
    dim="init",
    **metric_kwargs,
):
    """Computes the skill of a climatology forecast.

    Args:
        hind (xarray object): The initialized ensemble.
        verif (xarray object): control data, not needed
        metric (str): Metric name to apply at each lag for the persistence computation.
            Default: 'pearson_r'
        dim (str or list of str): dimension to apply metric over.
        add_attrs (bool): write climpred compute_persistence args to attrs.
            default: True
        ** metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        clim (xarray object): Results of climatology forecast with the input metric
            applied.
    """
    seasonality_str = OPTIONS["seasonality"]

    if isinstance(dim, str):
        dim = [dim]
    # Check that init is int, cftime, or datetime; convert ints or cftime to datetime.
    hind = convert_time_index(hind, "init", "hind[init]")
    verif = convert_time_index(verif, "time", "verif[time]")
    # Put this after `convert_time_index` since it assigns 'years' attribute if the
    # `init` dimension is a `float` or `int`.
    has_valid_lead_units(hind)

    # get metric/comparison function name, not the alias
    metric = METRIC_ALIASES.get(metric, metric)
    comparison = COMPARISON_ALIASES.get(comparison, comparison)

    comparison = get_comparison_class(comparison, ALL_COMPARISONS)
    metric = get_metric_class(metric, ALL_METRICS)

    if "iteration" in hind.dims:
        hind = hind.isel(iteration=0, drop=True)

    if comparison.hindcast:
        kind = "hindcast"
    else:
        kind = "perfect"

    if kind == "perfect":
        forecast, verif = comparison.function(hind, metric=metric)
        climatology_day = verif.groupby(f"init.{seasonality_str}").mean()
    else:
        forecast, verif = comparison.function(hind, verif, metric=metric)
        climatology_day = verif.groupby(f"time.{seasonality_str}").mean()

    climatology_day_forecast = climatology_day.sel(
        {seasonality_str: getattr(forecast.init.dt, seasonality_str)}, method="nearest"
    ).drop(seasonality_str)

    if kind == "hindcast":
        climatology_day_forecast = climatology_day_forecast.rename({"init": "time"})
    dim = _rename_dim(dim, climatology_day_forecast, verif)
    if metric.normalize:
        metric_kwargs["comparison"] = __e2c

    climatology_day_forecast, dim = _adapt_member_for_reference_forecast(
        climatology_day_forecast, verif, metric, comparison, dim
    )

    clim_skill = metric.function(
        climatology_day_forecast, verif, dim=dim, **metric_kwargs
    )
    if M2M_MEMBER_DIM in clim_skill.dims:
        clim_skill = clim_skill.mean(M2M_MEMBER_DIM)
    return clim_skill


@is_xarray([0, 1])
def compute_persistence(
    hind,
    verif,
    metric="pearson_r",
    alignment="same_verifs",
    add_attrs=True,
    dim="init",
    comparison="m2o",
    **metric_kwargs,
):
    """Computes the skill of a persistence forecast from a simulation.

    Args:
        hind (xarray object): The initialized ensemble.
        verif (xarray object): Verification data.
        metric (str): Metric name to apply at each lag for the persistence computation.
            Default: 'pearson_r'
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be based
            on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior to
            computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        dim (str or list of str): dimension to apply metric over.
        add_attrs (bool): write climpred compute_persistence args to attrs.
            default: True
        ** metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        pers (xarray object): Results of persistence forecast with the input metric
            applied.

    Reference:
        * Chapter 8 (Short-Term Climate Prediction) in Van den Dool, Huug.
          Empirical methods in short-term climate prediction.
          Oxford University Press, 2007.

    """
    if isinstance(dim, str):
        dim = [dim]
    # Check that init is int, cftime, or datetime; convert ints or cftime to datetime.
    hind = convert_time_index(hind, "init", "hind[init]")
    verif = convert_time_index(verif, "time", "verif[time]")
    # Put this after `convert_time_index` since it assigns 'years' attribute if the
    # `init` dimension is a `float` or `int`.
    has_valid_lead_units(hind)

    # get metric/comparison function name, not the alias
    metric = METRIC_ALIASES.get(metric, metric)
    comparison = COMPARISON_ALIASES.get(comparison, comparison)

    comparison = get_comparison_class(comparison, ALL_COMPARISONS)

    # get class metric(Metric)
    metric = get_metric_class(metric, ALL_METRICS)
    # If lead 0, need to make modifications to get proper persistence, since persistence
    # at lead 0 is == 1.
    if [0] in hind.lead.values:
        hind = hind.copy()
        with xr.set_options(keep_attrs=True):  # keeps lead.attrs['units']
            hind["lead"] = hind["lead"] + 1
        n, freq = get_lead_cftime_shift_args(hind.lead.attrs["units"], 1)
        # Shift backwards shift for lead zero.
        hind["init"] = shift_cftime_index(hind, "init", -1 * n, freq)
    # temporarily change `init` to `time` for comparison to verification data time.
    hind = hind.rename({"init": "time"})

    inits, verif_dates = return_inits_and_verif_dates(hind, verif, alignment=alignment)

    if metric.normalize:
        metric_kwargs["comparison"] = __e2c
    dim = _rename_dim(dim, hind, verif)

    plag = []
    for i in hind.lead.values:
        lforecast = verif.sel(time=inits[i])
        lverif = verif.sel(time=verif_dates[i])
        lforecast, dim = _adapt_member_for_reference_forecast(
            lforecast, lverif, metric, comparison, dim
        )
        lverif["time"] = lforecast["time"]
        # comparison expected for normalized metrics
        plag.append(metric.function(lforecast, lverif, dim=dim, **metric_kwargs))
    pers = xr.concat(plag, "lead")
    if "time" in pers:
        pers = pers.dropna(dim="time").rename({"time": "init"})
    pers["lead"] = hind.lead.values
    if add_attrs:
        pers = assign_attrs(
            pers,
            hind,
            function_name=inspect.stack()[0][3],
            alignment=alignment,
            metric=metric,
            metadata_dict=metric_kwargs,
        )
    return pers


@is_xarray([0, 1])
def compute_uninitialized(
    hind,
    uninit,
    verif,
    metric="pearson_r",
    comparison="e2o",
    dim="time",
    alignment="same_verifs",
    add_attrs=True,
    **metric_kwargs,
):
    """Verify an uninitialized ensemble against verification data.

    .. note::
        Based on Decadal Prediction protocol, this should only be computed for the
        first lag and then projected out to any further lags being analyzed.

    Args:
        hind (xarray object): Initialized ensemble.
        uninit (xarray object): Uninitialized ensemble.
        verif (xarray object): Verification data with some temporal overlap with the
            uninitialized ensemble.
        metric (str):
            Metric used in comparing the uninitialized ensemble with the verification
            data.
        comparison (str):
            How to compare the uninitialized ensemble to the verification data:
                * e2o : ensemble mean to verification data (Default)
                * m2o : each member to the verification data
        dim (str or list of str): dimension to apply metric over.
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be based
            on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior to
            computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        add_attrs (bool): write climpred compute args to attrs. default: True
        ** metric_kwargs (dict): additional keywords to be passed to metric

    Returns:
        u (xarray object): Results from comparison at the first lag.

    """
    if isinstance(dim, str):
        dim = [dim]
    # Check that init is int, cftime, or datetime; convert ints or cftime to datetime.
    hind = convert_time_index(hind, "init", "hind[init]")
    uninit = convert_time_index(uninit, "time", "uninit[time]")
    verif = convert_time_index(verif, "time", "verif[time]")
    has_valid_lead_units(hind)

    # get metric/comparison function name, not the alias
    metric = METRIC_ALIASES.get(metric, metric)
    comparison = COMPARISON_ALIASES.get(comparison, comparison)

    comparison = get_comparison_class(comparison, HINDCAST_COMPARISONS)
    metric = get_metric_class(metric, DETERMINISTIC_HINDCAST_METRICS)
    forecast, verif = comparison.function(uninit, verif, metric=metric)

    hind = hind.rename({"init": "time"})

    _, verif_dates = return_inits_and_verif_dates(hind, verif, alignment=alignment)

    if metric.normalize:
        metric_kwargs["comparison"] = comparison

    plag = []
    # TODO: `same_verifs` does not need to go through the loop, since it's a fixed
    # skill over all leads
    for i in hind["lead"].values:
        # Ensure that the uninitialized reference has all of the
        # dates for alignment.
        dates = list(set(forecast["time"].values) & set(verif_dates[i]))
        lforecast = forecast.sel(time=dates)
        lverif = verif.sel(time=dates)
        lforecast, dim = _adapt_member_for_reference_forecast(
            lforecast, lverif, metric, comparison, dim
        )
        lforecast["time"] = lverif["time"]
        # comparison expected for normalized metrics
        plag.append(metric.function(lforecast, lverif, dim=dim, **metric_kwargs))
    uninit_skill = xr.concat(plag, "lead")
    uninit_skill["lead"] = hind.lead.values

    # Attach climpred compute information to skill
    if add_attrs:
        uninit_skill = assign_attrs(
            uninit_skill,
            uninit,
            function_name=inspect.stack()[0][3],
            metric=metric,
            comparison=comparison,
            metadata_dict=metric_kwargs,
        )
    return uninit_skill



import numpy as np
import pytest

from climpred.constants import VALID_ALIGNMENTS
from climpred.reference import compute_uninitialized


@pytest.mark.parametrize("alignment", VALID_ALIGNMENTS)
def test_compute_uninitialized_alignment(
    hind_ds_initialized_1d, reconstruction_ds_1d, hist_ds_uninitialized_1d, alignment
):
    """Tests that compute_uninitialized works for various alignments."""
    res = (
        compute_uninitialized(
            hind_ds_initialized_1d,
            hist_ds_uninitialized_1d,
            reconstruction_ds_1d,
            metric="pr",
            comparison="e2o",
            alignment=alignment,
        )
        .isnull()
        .any()
    )
    for var in res.data_vars:
        assert not res[var]


def test_compute_uninitialized_same_verifs(
    hind_da_initialized_1d, reconstruction_da_1d, hist_da_uninitialized_1d
):
    """Tests that uninitialized skill is same at all leads for `same_verifs`
    alignment."""
    res = compute_uninitialized(
        hind_da_initialized_1d,
        hist_da_uninitialized_1d,
        reconstruction_da_1d,
        metric="pr",
        comparison="e2o",
        alignment="same_verifs",
    )
    assert ((res - res[0]) == 0).all()
